/* #pragma ident	"@(#)rsvp_cbq.c	1.42	97/08/06 SMI" */
/*
 * @(#) $Id: rsvp_cbq.c,v 0.18 1999/12/28 11:56:46 kjc Exp kjc $
 */

/************************ rsvp_kern.c  *******************************
 *                                                                   *
 *  Interface between RSVP and traffic control modules in kernel.    *
 *     This module interfaces to CBQ, the Sun/UCL/LBL kernel.        *
 *                                                                   *
 *********************************************************************/

/*******************************************************************

  Copyright (c) 1998 by the University of Southern California
  All rights reserved.

  Permission to use, copy, modify, and distribute this software and its
  documentation in source and binary forms for any purpose and without
  fee is hereby granted, provided that both the above copyright notice
  and this permission notice appear in all copies. and that any
  documentation, advertising materials, and other materials related to
  such distribution and use acknowledge that the software was developed
  in part by the University of Southern California, Information
  Sciences Institute.  The name of the University may not be used to
  endorse or promote products derived from this software without
  specific prior written permission.

  THE UNIVERSITY OF SOUTHERN CALIFORNIA makes no representations about
  the suitability of this software for any purpose.  THIS SOFTWARE IS
  PROVIDED "AS IS" AND WITHOUT ANY EXPRESS OR IMPLIED WARRANTIES,
  INCLUDING, WITHOUT LIMITATION, THE IMPLIED WARRANTIES OF
  MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.

  Other copyrights might apply to parts of this software and are so
  noted when applicable.

********************************************************************/

/*
 * Copyright (c) Sun Microsystems, Inc. 1993-1998 All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 *
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *      This product includes software developed by the SMCC Technology
 *      Development Group at Sun Microsystems, Inc.
 *
 * 4. The name of the Sun Microsystems, Inc nor may not be used to endorse or
 *      promote products derived from this software without specific prior
 *      written permission.
 *
 * SUN MICROSYSTEMS DOES NOT CLAIM MERCHANTABILITY OF THIS SOFTWARE OR THE
 * SUITABILITY OF THIS SOFTWARE FOR ANY PARTICULAR PURPOSE.  The software is
 * provided "as is" without express or implied warranty of any kind.
 *  
 * These notices must be retained in any copies of any part of this software.
 */

#include <sys/param.h>
#include <stddef.h>

#ifdef CBQDAEMON
#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>
#include <ctype.h>
#include <errno.h>
#include <err.h>
#include <syslog.h>
#include <sys/socket.h>
#include <net/if.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#else
#include "rsvp_daemon.h"
#include "rsvp_TCif.h"		/* Adaptation interface */
#endif
#include <altq/altq.h>
#include <altq/altq_cbq.h>

#ifdef CBQDAEMON
#include "rsvp_cbq.h"
#else
#include "tc_cbq.h"
#endif
#include <fcntl.h>
#include <stdio.h>
#include <string.h>
#include <sys/ioctl.h>
#include <limits.h>

/*
 * Define local Macros
 */
#define CTL_PBANDWIDTH	2
#define NS_PER_MS	(1000000.0)
#define NS_PER_SEC	(NS_PER_MS*1000.0)
#define RM_FILTER_GAIN	5

#define	CBQ_DEVICE	"/dev/altq/cbq"

#define ADMIT_FLOW_OK		0
#define ADMIT_FLOW_NOSVC	1
#define ADMIT_FLOW_NOBW		2
#define ADMIT_FLOW_SYSERR	3

#ifndef CBQDAEMON

typedef struct {
	int		in_if;		/* interface to which resv applies */
	cbq_flowinfo_t 	*cbq_handle;	/* cbq class handle */
	int		flags;
	u_short		nfilters;	/* Count filters for SE */
	SENDER_TSPEC	*s_tspec;	/* ptr to SENDER_TSPEC object */
	FLOWSPEC	*spec;		/* ptr to FLOWSPEC object */
	FILTER_SPEC	*filter;	/* ptr to FILTER_SPEC object */
	struct sockaddr_in	dest;
} tc_t;

#endif

/* 
 * External declarations (don't include math.h, log() conflicts)
 */
extern char           *fmt_spec();
extern double          pow(double, double);
extern double          fabs(double);

/* 
 * Forward references
 */
static int cbq_clear(char *ifname);
static cbq_ifinfo_t	*ifname2ifinfo(char *ifname);
static cbq_classinfo_t	*clname2clinfo(cbq_ifinfo_t *ifinfo, char *clname);
static int check_bandwidth(cbq_ifinfo_t *ifinfo, cbq_classinfo_t *parent);
static void pr_filter(struct flow_filter * filt);
static int cbq_add_ctl_filters(char *ifname, char *clname);
#ifndef CBQDAEMON
static int rsvp2cbq_filter(FilterSpec *filtp, Session *dest,
			   struct flow_filter *fl);
static int admit_send_flow(tc_t *, FLOWSPEC *, SENDER_TSPEC *, int);
static void unadmit_send_flow(tc_t *);
static cbq_flowinfo_t *cbq_addflow(tc_t *);
static int cbq_delflow(cbq_flowinfo_t * flow_info);
static cbq_flowinfo_t *cbq_modifyflow(tc_t *);
static cbq_fltrinfo_t *cbq_addfilter(cbq_flowinfo_t * flinfo,
				     struct flow_filter * cfilter);
static int cbq_delfilter_all(cbq_flowinfo_t * flow_info);
static int cbq_delfilter(cbq_fltrinfo_t *);
extern Object_header	*copy_object(Object_header *);
#endif /* !CBQDAEMON */

/*
 * Static globals
 */

static cbq_ifinfo_t	*cbq_iflist = NULL;
static int		cbqfd = -1;

/*********************************************************************
 *  Following are CBQ-specific and are
 *  are called from above by RSVP.
 *
 *********************************************************************/

static cbq_ifinfo_t   *
ifname2ifinfo(char *ifname)
{
	cbq_ifinfo_t   *ifinfo;

	for (ifinfo = cbq_iflist; ifinfo; ifinfo = ifinfo->next)
		if (strcmp(ifinfo->ifname, ifname) == 0)
			return (ifinfo);
	return (NULL);
}

static cbq_classinfo_t	*
clname2clinfo(cbq_ifinfo_t *ifinfo, char *clname)
{
	cbq_classinfo_t	*clinfo;
	
	for (clinfo = ifinfo->cllist; clinfo; clinfo = clinfo->next)
		if (strcmp(clinfo->clname, clname) == 0)
			return (clinfo);
	return (NULL);
}

static int 
check_bandwidth(cbq_ifinfo_t *ifinfo, cbq_classinfo_t *parent)
{
	cbq_classinfo_t	*clinfo;
	u_int bandwidth = 0;

	if (parent == NULL)
		return (0);
	for (clinfo = ifinfo->cllist; clinfo; clinfo = clinfo->next)
		if (clinfo->parent_cl == parent)
			bandwidth += clinfo->acinfo.bandwidth;

	if (bandwidth > parent->acinfo.bandwidth) {
		log(LOG_WARNING, 0,
		    "warning: children's total bandwidth (%dKbps) exceeds their parent's (%dKbps) on interface %s\n",
		    bandwidth/1000, parent->acinfo.bandwidth/1000,
		    ifinfo->ifname);
		return (-1);
	}

	return (0);
}

int
cbq_init()
{
	if (cbqfd >= 0)
	    close(cbqfd);

	cbq_cmds_init();

	if ((cbqfd = open(CBQ_DEVICE, O_RDWR)) < 0) {
		log(LOG_ERR, errno, "CBQ open");
		return (-1);
	}

	/* read config file and initialize CBQ */
	return cbq_config();
}

int
cbq_add_if(char *ifname, u_int bandwidth, int is_sched, int is_wrr,
	   int efficient)
{
	cbq_ifinfo_t   *ifinfo;
	struct cbq_interface	cbqif;

	if ((ifinfo = calloc(1, sizeof(cbq_ifinfo_t))) == NULL)
		return (-1);
	ifinfo->ifname = strdup(ifname);
	ifinfo->cllist = NULL;
	ifinfo->bandwidth = bandwidth;
	ifinfo->nsPerByte =
		((1.0 / (double)ifinfo->bandwidth) * NS_PER_SEC * 8);
	ifinfo->use_cbq = is_sched && (cbqfd >= 0);
	ifinfo->is_wrr = is_wrr;
	ifinfo->is_efficient = efficient;
	ifinfo->enabled = 0;
	ifinfo->ifmtu = 0;

	/* Link the interface info structure */
	ifinfo->next = cbq_iflist;
	cbq_iflist = ifinfo;

	/* get mtu of the interface for default packet size */
	do {
		int s;
		struct ifreq ifr;
#ifdef __OpenBSD__
		struct if_data ifdata;
#endif

		if ((s = socket(AF_INET, SOCK_DGRAM, 0)) < 0)
			goto bad;
		strncpy(ifr.ifr_name, ifname, sizeof ifr.ifr_name);
#ifdef __OpenBSD__
		ifr.ifr_data = (caddr_t)&ifdata;
		if (ioctl(s, SIOCGIFDATA, (caddr_t)&ifr) < 0)
			goto bad;
		ifinfo->ifmtu = ifdata.ifi_mtu;
#else
		if (ioctl(s, SIOCGIFMTU, (caddr_t)&ifr) < 0)
			goto bad;
		ifinfo->ifmtu = ifr.ifr_mtu;
#endif
	bad:
		if (ifinfo->ifmtu == 0)
			ifinfo->ifmtu = 512;	/* XXX */
		if (s >= 0)
			close(s);
	} while (0);

	/* attach cbq to the interface */
	memset((void *) &cbqif, 0, sizeof(cbqif));
	strncpy(cbqif.cbq_ifacename, ifname, IFNAMSIZ);
	if (ifinfo->use_cbq &&
	    ioctl(cbqfd, CBQ_IF_ATTACH, &cbqif) < 0) {
		log(LOG_ERR, errno, "ioctl CBQ_IF_ATTACH");
		cbq_iflist = cbq_iflist->next;
		goto err_ret;
	}

	if (cbq_disable(ifname) || cbq_clear(ifname)) {
		cbq_iflist = cbq_iflist->next;
		goto err_ret;
	}
	return (0);

err_ret:
	if (ifinfo) {
		if (ifinfo->ifname)
			free(ifinfo->ifname);
		free(ifinfo);
	}
	return (-1);
}

int
cbq_enableall()
{
	cbq_ifinfo_t   *ifinfo;

	for (ifinfo = cbq_iflist; ifinfo; ifinfo = ifinfo->next)
		if (cbq_enable(ifinfo->ifname) < 0)
			return (-1);
	return (0);
}

int
cbq_enable(char *ifname)
{
	cbq_ifinfo_t   *ifinfo;
	struct cbq_interface cbqif;

	if ((ifinfo = ifname2ifinfo(ifname)) == NULL)
		return (-1);

	/*
	 * Enable the interface
	 */
	memset((void *) &cbqif, 0, sizeof(cbqif));
	strncpy(cbqif.cbq_ifacename, ifname, IFNAMSIZ);
	if (ifinfo->use_cbq &&
	    ioctl(cbqfd, CBQ_ENABLE, &cbqif) < 0) {
		log(LOG_ERR, errno, "ioctl CBQ_ENABLE");
		return (-1);
	}
	ifinfo->enabled = 1;

	log(LOG_INFO, 0, "CBQ enabled on interface %s (mtu:%d)\n",
	    ifname, ifinfo->ifmtu);

	return (0);
}

int
cbq_disableall()
{
	cbq_ifinfo_t   *ifinfo;
	int	error = 0;
	
	for (ifinfo = cbq_iflist; ifinfo; ifinfo = ifinfo->next)
		if (cbq_disable(ifinfo->ifname) < 0)
			error = -1;
	return (error);
}

int
cbq_disable(char *ifname)
{
	struct cbq_interface cbqif;
	cbq_ifinfo_t   *ifinfo;

	if ((ifinfo = ifname2ifinfo(ifname)) == NULL)
		return (-1);

	/* Disable the interface */
	memset((void *) &cbqif, 0, sizeof(cbqif));
	strncpy(cbqif.cbq_ifacename, ifname, IFNAMSIZ);
	if (ifinfo->use_cbq &&
	    ioctl(cbqfd, CBQ_DISABLE, &cbqif) < 0) {
		log(LOG_ERR, errno, "ioctl CBQ_DISABLE");
		return (-1);
	}
	ifinfo->enabled = 0;
	return (0);
}

int cbq_is_enabled(char *ifname)
{
	cbq_ifinfo_t   *ifinfo;

	if ((ifinfo = ifname2ifinfo(ifname)) == NULL)
		return (0);
	return (ifinfo->enabled);
}

static int
cbq_class_spec(cbq_ifinfo_t * ifinfo, u_long parent_class,
	u_long borrow_class, u_int pri, int flags,
	u_int bandwidth, u_int maxdelay, u_int maxburst,
	u_int minburst, u_int av_pkt_size, u_int max_pkt_size,
	cbq_class_spec_t *cl_spec)
{
	double          maxq, maxidle_s, maxidle, minidle,
			offtime, nsPerByte, ptime, cptime;
	double		z = (double)(1 << RM_FILTER_GAIN);
	double          g = (1.0 - 1.0 / z);
	double          f;
 	double		gton;
 	double		gtom;
	double		maxrate;

 	/* Compute other class parameters */
	if (bandwidth == 0)
		f = 0.0001;	/* small enough? */
	else
		f = ((double) bandwidth / (double) ifinfo->bandwidth);

	if (av_pkt_size == 0) {	/* use default */
		av_pkt_size = ifinfo->ifmtu;
		if (av_pkt_size > MCLBYTES)	/* do what TCP does */
			av_pkt_size &= ~MCLBYTES;
	}
	else if (av_pkt_size > ifinfo->ifmtu)
		av_pkt_size = ifinfo->ifmtu;
	if (max_pkt_size == 0)	/* use default */
		max_pkt_size = ifinfo->ifmtu;
	else if (max_pkt_size > ifinfo->ifmtu)
		max_pkt_size = ifinfo->ifmtu;

        nsPerByte = ifinfo->nsPerByte / f;
	ptime = (double) av_pkt_size * (double)ifinfo->nsPerByte;
	maxrate = f * ((double)ifinfo->bandwidth / 8.0);
	cptime = ptime * (1.0 - f) / f;
#if 1 /* ALTQ */
	if (nsPerByte * (double)max_pkt_size > (double)INT_MAX) {
		/*
		 * this causes integer overflow in kernel!
		 * (bandwidth < 6Kbps when max_pkt_size=1500)
		 */
		if (bandwidth != 0)
			log(LOG_WARNING, 0, "warning: class is too slow!!\n");
		nsPerByte = (double)(INT_MAX / max_pkt_size);
	}
#endif
	if (maxburst == 0) {  /* use default */
		if (cptime > 10.0 * NS_PER_MS)
			maxburst = 4;
		else
			maxburst = 16;
	}
	if (minburst == 0)  /* use default */
		minburst = 2;
	if (minburst > maxburst)
		minburst = maxburst;

	if (IsDebug(DEBUG_CBQ)) {
		int packet_time;
		log(LOG_DEBUG, 0,
		    "cbq_flowspec: maxburst=%d,minburst=%d,pkt_size=%d\n",
		    maxburst, minburst, av_pkt_size);
		log(LOG_DEBUG, 0,
		    "  nsPerByte=%.2f ns, link's nsPerByte=%.2f, f=%.3f\n",
		    nsPerByte, ifinfo->nsPerByte, f);
		packet_time = av_pkt_size * (int)nsPerByte / 1000;
		log(LOG_DEBUG, 0,
		    "  packet time=%d [us]\n", packet_time);
		if (maxburst * packet_time < 20000) {
			log(LOG_WARNING, 0,
			  "warning: maxburst smaller than timer granularity!\n");
			log(LOG_WARNING, 0,
			    "         maxburst=%d, packet_time=%d [us]\n",
			    maxburst, packet_time);
		}
	}
 	gton = pow(g, (double)maxburst);
 	gtom = pow(g, (double)(minburst-1));
	maxidle = ((1.0 / f - 1.0) * ((1.0 - gton) / gton));
	maxidle_s = (1.0 - g);
	if (maxidle > maxidle_s)
		maxidle = ptime * maxidle;
	else
		maxidle = ptime * maxidle_s;
	if (IsDebug(DEBUG_CBQ))
		log(LOG_DEBUG, 0, "  maxidle=%.2f us\n", maxidle/1000.0);
	if (minburst)
		offtime = cptime * (1.0 + 1.0/(1.0 - g) * (1.0 - gtom) / gtom);
	else
		offtime = cptime;
	minidle = -((double)max_pkt_size * (double)nsPerByte);
	if (IsDebug(DEBUG_CBQ))
		log(LOG_DEBUG, 0, "  offtime=%.2f us minidle=%.2f us\n",
		    offtime/1000.0, minidle/1000.0);

	maxidle = ((maxidle * 8.0) / nsPerByte) * pow(2, RM_FILTER_GAIN);
#if 1 /* ALTQ */
	/* also scale offtime and minidle */
 	offtime = (offtime * 8.0) / nsPerByte * pow(2, RM_FILTER_GAIN);
	minidle = ((minidle * 8.0) / nsPerByte) * pow(2, RM_FILTER_GAIN);
#endif
	maxidle = maxidle / 1000.0;
	offtime = offtime / 1000.0;
	minidle = minidle / 1000.0;
	/* adjust queue size when maxdelay is specified.
	   queue size should be relative to its share */
	if (maxdelay == 0) {
		if (flags & (CBQCLF_RED|CBQCLF_RIO))
			maxq = 60.0;
		else
			maxq = 30.0;
	}
	else {
		maxq = ((double) maxdelay * NS_PER_MS) / (nsPerByte * av_pkt_size);
		if (maxq < 4) {
			log(LOG_WARNING, 0,
			    "warning: maxq (%d) is too small. set to %d\n",
			    (int)maxq, 4);
			maxq = 4;
		}
	}
	if (bandwidth == 0 && borrow_class == NULL_CLASS_HANDLE)
		/* filter out this class by setting queue size to zero */
		maxq = 0;
	if (IsDebug(DEBUG_CBQ)) {
		if ((u_int)maxq < maxburst)
			log(LOG_WARNING, 0,
			   "warning: maxq (%d) is smaller than maxburst(%d)\n",
			    (int)maxq, maxburst);
		else if (maxq > 100.0)
			log(LOG_WARNING, 0,
			    "warning: maxq %d too large\n", (int)maxq);
		log(LOG_DEBUG, 0, "  maxq=%d\n", (int)maxq);
	}

	if (parent_class == NULL_CLASS_HANDLE) {
		if ((flags & CBQCLF_ROOTCLASS) == 0)
			flags |= CBQCLF_ROOTCLASS;
		if (ifinfo->is_wrr)
			flags |= CBQCLF_WRR;
		if (ifinfo->is_efficient)
			flags |= CBQCLF_EFFICIENT;
	}

	memset((void *)cl_spec, 0, sizeof(cbq_class_spec_t));
	cl_spec->priority = pri;
	cl_spec->nano_sec_per_byte = (u_int) nsPerByte;
	cl_spec->maxq = (u_int) maxq;
	cl_spec->maxidle = (u_int) fabs(maxidle);
	cl_spec->minidle = (int)minidle;
	cl_spec->offtime = (u_int) fabs(offtime);

	cl_spec->parent_class_handle = parent_class;
	cl_spec->borrow_class_handle = borrow_class;

	cl_spec->pktsize = av_pkt_size;
	cl_spec->flags = flags;

	return (0);
}

static u_long
cbq_create_class(cbq_ifinfo_t * ifinfo, u_long parent_class,
	u_long borrow_class, u_int pri, int flags,
	u_int bandwidth, u_int maxdelay, u_int maxburst,
	u_int minburst, u_int av_pkt_size, u_int max_pkt_size)
{
	struct cbq_add_class cl_add;

	memset((void *) &cl_add, 0, sizeof(cl_add));
	strncpy(cl_add.cbq_iface.cbq_ifacename, ifinfo->ifname,IFNAMSIZ);

	if (cbq_class_spec(ifinfo, parent_class, borrow_class, pri, flags,
			  bandwidth, maxdelay, maxburst, minburst,
			  av_pkt_size, max_pkt_size,
			  &cl_add.cbq_class) < 0) {
		return (NULL_CLASS_HANDLE);
	}

	if (ifinfo->use_cbq &&
	    ioctl(cbqfd, CBQ_ADD_CLASS, &cl_add) < 0) {
		log(LOG_ERR, errno, "ioctl CBQ_ADD_CLASS");
		return (NULL_CLASS_HANDLE);
	}

	return cl_add.cbq_class_handle;
}

#ifndef CBQDAEMON
/*
 * modify the bandwidth of the specified flow
 *	the following parameters can't be changed:
 *		parent_class, borrow_class, pri, flags
 */
static u_long
cbq_modify_class(cbq_ifinfo_t * ifinfo, u_long chandle, u_long parent_class,
	u_long borrow_class, u_int pri, int flags,
	u_int bandwidth, u_int maxdelay, u_int maxburst,
	u_int minburst, u_int av_pkt_size, u_int max_pkt_size)
{
	struct cbq_modify_class cl_modify;

	memset((void *) &cl_modify, 0, sizeof(cl_modify));
	strncpy(cl_modify.cbq_iface.cbq_ifacename,
		ifinfo->ifname, IFNAMSIZ);
	cl_modify.cbq_class_handle = chandle;

	if (cbq_class_spec(ifinfo, parent_class, borrow_class, pri, flags,
			  bandwidth, maxdelay, maxburst, minburst,
			  av_pkt_size, max_pkt_size,
			  &cl_modify.cbq_class) < 0) {
		return (NULL_CLASS_HANDLE);
	}

	if (ifinfo->use_cbq &&
	    ioctl(cbqfd, CBQ_MODIFY_CLASS, &cl_modify) < 0) {
		log(LOG_ERR, errno, "ioctl CBQ_MODIFY_CLASS");
		return (NULL_CLASS_HANDLE);
	}

	return cl_modify.cbq_class_handle;
}
#endif /* !CBQDAEMON */

int
cbq_add_static_class(char *ifname, char *clname, char *parent_class,
     	char *borrow_class, u_int pri, int flags, u_int pbandwidth,
	u_int maxdelay, u_int maxburst, u_int minburst, u_int av_pkt_size,
	u_int max_pkt_size, cbq_tos_t admission_type,
	cbq_ol_t overlimit_action, u_int exactbandwidth)
{
	cbq_ifinfo_t   *ifinfo;
	cbq_classinfo_t *parent_cl = NULL, *borrow_cl = NULL,
			*ctl_cl = NULL, *new_cl;
	u_long          cl_handle = NULL_CLASS_HANDLE;
	u_int            bandwidth;

	if ((ifinfo = ifname2ifinfo(ifname)) == NULL)
		return (-1);

	if (strcmp("ctl_class", clname) == 0)
		flags |= CBQCLF_CTLCLASS;
	else if (flags & CBQCLF_CTLCLASS)
		clname = "ctl_class";

	/* Get parent info */
	if (parent_class)
		if ((parent_cl = clname2clinfo(ifinfo, parent_class)) == NULL)
			return (-1);
	
	/* Get borrow info */
	if (borrow_class) {
		if ((borrow_cl = clname2clinfo(ifinfo, borrow_class)) == NULL)
			return (-1);
		if (borrow_cl != parent_cl) {
		    log(LOG_ERR, 0, "borrow class should be parent!\n");
		    return (-1);
		}
	}

	/* Create class */
	if (flags & CBQCLF_DEFCLASS) {
		/*
		 * if this is a default class and no ctl_class is defined,
		 * create a ctl_class.
		 */
		ctl_cl = clname2clinfo(ifinfo, "ctl_class");
		if (ctl_cl == NULL) {
			/* reserve bandwidth for ctl_class */
			if (exactbandwidth != 0)
				exactbandwidth -= ifinfo->bandwidth / 100
					* CTL_PBANDWIDTH;
			else
				pbandwidth -= CTL_PBANDWIDTH;
		}
	}

	if (exactbandwidth != 0) {
		/* if exactbandwidth is specified, it overrides pbandwidth */
		bandwidth = exactbandwidth;
	}
	else
		bandwidth = ifinfo->bandwidth / 100 * pbandwidth;

	/* check bandwidth is lower than its parent */
	if (parent_cl != NULL && bandwidth > parent_cl->acinfo.bandwidth) {
		log(LOG_ERR, 0,
		    "%s: bandwidth (%dKbps) greater then it parent's (%dKbps)\n",
		    clname, bandwidth/1000, parent_cl->acinfo.bandwidth/1000);
		return (-1);
	}

	/* if average paket size isn't specified, set if mtu. */
	if (av_pkt_size == 0) {	/* use default */
		av_pkt_size = ifinfo->ifmtu;
		if (av_pkt_size > MCLBYTES)	/* do what TCP does */
			av_pkt_size &= ~MCLBYTES;
	}
	else if (av_pkt_size > ifinfo->ifmtu)
		av_pkt_size = ifinfo->ifmtu;

	if (max_pkt_size == 0)	/* use default */
		max_pkt_size = ifinfo->ifmtu;
	else if (max_pkt_size > ifinfo->ifmtu)
		max_pkt_size = ifinfo->ifmtu;

	/* instantiate class descriptor and init */
	if ((new_cl = calloc(1, sizeof(cbq_classinfo_t))) == NULL)
		return (-1);

	cl_handle = cbq_create_class(ifinfo,
		(parent_cl ? parent_cl->handle : NULL_CLASS_HANDLE),
		(borrow_cl ? borrow_cl->handle : NULL_CLASS_HANDLE),
	      	pri, flags, bandwidth, maxdelay,
		maxburst, minburst, av_pkt_size, max_pkt_size);
	if (cl_handle == NULL_CLASS_HANDLE)
		goto err_ret;

	new_cl->clname = strdup(clname);
	new_cl->handle = cl_handle;
	new_cl->parent_cl = parent_cl;
	new_cl->borrow_cl = borrow_cl;
	new_cl->ifinfo = ifinfo;

	/* Set pointers if a guranteed or predictive non-leaf class */
	switch (admission_type) {
	case CBQ_QOS_CNTR_LOAD:
	case CBQ_QOS_GUARANTEED:
#if 1
	case CBQ_QOS_PREDICTIVE:
	case CBQ_QOS_CNTR_DELAY:
#endif
		if (ifinfo->resv_class) {
			log(LOG_ERR, 0,
			    "%s: duplicate resv meta class\n", clname);
			goto err_ret;
		}
		ifinfo->resv_class = new_cl;
		/* fall into... */
	case CBQ_QOS_NONE:
		new_cl->acinfo.type = admission_type;
		new_cl->acinfo.pri = pri;
		new_cl->acinfo.bandwidth = bandwidth;
		new_cl->acinfo.allocated = 0;
		new_cl->acinfo.maxdelay = maxdelay;
		new_cl->acinfo.maxburst = maxburst;
		new_cl->acinfo.minburst = minburst;
		new_cl->acinfo.av_pkt_size = av_pkt_size;
		new_cl->acinfo.max_pkt_size = max_pkt_size;
		new_cl->acinfo.flags = flags;
		break;
	default:
		goto err_ret;
	}

	/* link classinfo in list */
	new_cl->next = ifinfo->cllist;
	ifinfo->cllist = new_cl;

	/*
	 * If were adding the default traffic class and no ctl_class is
	 * defined, also add the ctl traffic class.  This is for RSVP,
	 * and IGMP packets.
	 */
	if ((flags & CBQCLF_DEFCLASS) && ctl_cl == NULL) {
		if (cbq_add_static_class(ifname, "ctl_class", parent_class,
			borrow_class, 6, CBQCLF_CTLCLASS, CTL_PBANDWIDTH,
			maxdelay, maxburst, minburst, av_pkt_size,
			max_pkt_size, admission_type, overlimit_action, 0) < 0) {
			log(LOG_ERR, errno,
			    "can't create ctl_class");
		}
	}

	/*
	 * if this is a ctl class, add the default filters for backward
	 * compatibility
	 */
	if (flags & CBQCLF_CTLCLASS)
		cbq_add_ctl_filters(ifname, clname);

	check_bandwidth(ifinfo, new_cl->parent_cl);

	return (0);

err_ret:
	if (new_cl) {
		if (new_cl->clname)
			free(new_cl->clname);
		free(new_cl);
	}
	return (-1);
}

int
cbq_delete_static_class(char *ifname, char *clname)
{
	cbq_ifinfo_t   *ifinfo;
	cbq_classinfo_t *clinfo, *clprev;
	struct cbq_delete_class class_delete;

	if ((ifinfo = ifname2ifinfo(ifname)) == NULL)
		return (-1);

	/* remove class info from the interface */
	if ((clinfo = clprev = ifinfo->cllist) == NULL)
		return (-1);
	if (strcmp(clinfo->clname, clname) == 0) {
		ifinfo->cllist = clinfo->next;
	}
	else while ((clinfo = clprev->next) != NULL) {
		if (strcmp(clinfo->clname, clname) == 0) {
			clprev->next = clinfo->next;
			break;
		}
		clprev = clinfo;
	}

#ifndef CBQDAEMON
	/* delete flowinfo associated with this class */
	while (clinfo->acinfo.flow_list != NULL)
		if (cbq_delflow(clinfo->acinfo.flow_list) < 0) {
			log(LOG_ERR, errno, "can't delete flow");
			break;
		}
#endif

	/* delete class from kernel */
	if (clinfo->handle != NULL_CLASS_HANDLE) {
		memset((void *) &class_delete, 0, sizeof(class_delete));
		strncpy(class_delete.cbq_iface.cbq_ifacename,
			ifname, IFNAMSIZ);
		class_delete.cbq_class_handle = clinfo->handle;
		if (ifinfo->use_cbq &&
		    ioctl(cbqfd, CBQ_DEL_CLASS, &class_delete) < 0) {
			log(LOG_ERR, errno, "ioctl CBQ_DEL_CLASS");
			return (-1);
		}
	}

	if (clinfo->clname)
		free(clinfo->clname);
	free(clinfo);
	return (0);
}

int
cbq_add_static_filter(char *ifname, char *clname, struct flow_filter *filt)
{
	cbq_ifinfo_t   *ifinfo;
	cbq_classinfo_t *clinfo = NULL;
	struct cbq_add_filter fltr_add;

	if ((ifinfo = ifname2ifinfo(ifname)) == NULL)
		return (-1);

	if ((clinfo = clname2clinfo(ifinfo, clname)) == NULL)
		return (-1);

	memset((void *) &fltr_add, 0, sizeof(fltr_add));
	strncpy(fltr_add.cbq_iface.cbq_ifacename, ifname, IFNAMSIZ);
	fltr_add.cbq_class_handle = clinfo->handle;
	fltr_add.cbq_filter = *filt;

	if (IsDebug(DEBUG_CBQ)) {
		log(LOG_DEBUG, 0, "%s: CBQ_ADD_FILTER, class.%lx\n",
		    ifname, fltr_add.cbq_class_handle);
		pr_filter(&fltr_add.cbq_filter);
	}

	if (ifinfo->use_cbq &&
	    ioctl(cbqfd, CBQ_ADD_FILTER, &fltr_add) < 0) {
		log(LOG_ERR, errno, "ioctl CBQ_ADD_FILTER");
		return (-1);
	}
	return (0);
}

/*
 * add the default filters for ctl_class (for backward compatibility).
 */
#ifndef IPPROTO_RSVP
#define IPPROTO_RSVP		46
#endif

static int cbq_add_ctl_filters(ifname, clname)
	char *ifname, *clname;
{
	struct flow_filter	sfilt;
	u_int8_t ctl_protos[3] = {IPPROTO_ICMP, IPPROTO_IGMP, IPPROTO_RSVP};
#ifdef USE_IPV6
	struct flow_filter6	sfilt6;
	u_int8_t ctl6_protos[3] = {IPPROTO_ICMPV6, IPPROTO_IGMP, IPPROTO_RSVP};
#endif
	int ret, i;

	for (i = 0; i < sizeof(ctl_protos); i++) {
		memset(&sfilt, 0, sizeof(sfilt));
		sfilt.ff_flow.fi_family = AF_INET;
		sfilt.ff_flow.fi_proto = ctl_protos[i];

		ret = cbq_add_static_filter(ifname, clname, &sfilt);
		if (ret) {
			log(LOG_ERR, 0,
			    "can't add ctl class filter on interface '%s'\n",
			    ifname);
			return (-1);
		}
	}

#ifdef USE_IPV6
	for (i = 0; i < sizeof(ctl6_protos); i++) {
		memset(&sfilt6, 0, sizeof(sfilt6));
		sfilt6.ff_flow6.fi6_family = AF_INET6;
		sfilt6.ff_flow6.fi6_proto = ctl6_protos[i];

		ret = cbq_add_static_filter(ifname, clname,
					    (struct flow_filter *)&sfilt6);
		if (ret) {
			log(LOG_ERR, 0,
			    "can't add ctl class filter on interface '%s'\n",
			    ifname);
			return (-1);
		}
	}
#endif
	return (0);
}

	
static int
cbq_clear(char *ifname)
{
	struct cbq_interface 	cbqif;
	cbq_ifinfo_t  		*ifinfo;

	if ((ifinfo = ifname2ifinfo(ifname)) == NULL)
		return (-1);
		
	/* clear the interface */
	memset((void *) &cbqif, 0, sizeof(cbqif));
	strncpy(cbqif.cbq_ifacename, ifname, IFNAMSIZ);
	if (ifinfo->use_cbq &&
	    ioctl(cbqfd, CBQ_CLEAR_HIERARCHY, &cbqif) < 0) {
		log(LOG_ERR, errno, "ioctl CBQ_CLEAR_HIERARCHY");
		return (-1);
	}
	return (0);
}

int
cbq_destroyall(void)
{
	cbq_ifinfo_t		*ifinfo, *ifnext;
	cbq_classinfo_t		*clinfo, *clnext;
	struct cbq_interface	cbqif;

	if (cbqfd < 0) {
		log(LOG_ERR, 0, "cbq_destroyall: CBQ device is not opened\n");
		return (-1);
	}

	for (ifinfo = cbq_iflist; ifinfo; ifinfo = ifnext) {

		(void) cbq_disable(ifinfo->ifname);
		
		/* free all classes */
		for (clinfo = ifinfo->cllist; clinfo; clinfo = clnext) {
#ifndef CBQDAEMON
			/* delete flowinfo associated with this class */
			while (clinfo->acinfo.flow_list != NULL)
				if (cbq_delflow(clinfo->acinfo.flow_list) < 0) {
					log(LOG_ERR, errno,
					    "can't delete flow");
					break;
				}
#endif
			clnext = clinfo->next;
			if (clinfo->clname)
				free(clinfo->clname);
			free(clinfo);
		}

		(void)cbq_clear(ifinfo->ifname);

		memset((void *) &cbqif, 0, sizeof(cbqif));
		strncpy(cbqif.cbq_ifacename, ifinfo->ifname, IFNAMSIZ);
			
		if (ifinfo->use_cbq &&
		    ioctl(cbqfd, CBQ_IF_DETACH, &cbqif) < 0)
			log(LOG_ERR, errno, "ioctl CBQ_IF_DETACH");

		ifnext = ifinfo->next;
		if (ifinfo->ifname)
			free(ifinfo->ifname);
		free(ifinfo);
	}

	cbq_iflist = NULL;
	if (cbqfd >= 0 && close(cbqfd) < 0)
		log(LOG_ERR, errno, "close cbq");
	cbqfd = -1;
	return (0);
}

static void
pr_filter(struct flow_filter *filt)
{
	if (filt->ff_flow.fi_family == AF_INET) {
		struct in_addr in_addr;

		in_addr.s_addr = filt->ff_flow.fi_dst.s_addr;
		log(LOG_DEBUG, 0,
		    "Filter Dest Addr: %s (mask 0x%x) Port: %d\n",
		    inet_ntoa(in_addr), ntoh32(filt->ff_mask.mask_dst.s_addr),
		    ntoh16(filt->ff_flow.fi_dport));
		in_addr.s_addr = filt->ff_flow.fi_src.s_addr;
		log(LOG_DEBUG, 0,
		    "Filter Src Addr: %s (mask 0x%x) Port: %d\n",
		    inet_ntoa(in_addr), ntoh32(filt->ff_mask.mask_src.s_addr),
		    ntoh16(filt->ff_flow.fi_sport));
		log(LOG_DEBUG, 0, "Filter Protocol: %d TOS 0x%x (mask 0x%x)\n",
		    filt->ff_flow.fi_proto, filt->ff_flow.fi_tos,
		    filt->ff_mask.mask_tos);
	}
#ifdef INET6
	else if (filt->ff_flow.fi_family == AF_INET6) {
		char str1[INET6_ADDRSTRLEN], str2[INET6_ADDRSTRLEN];
		struct flow_filter6 *sfilt6 = (struct flow_filter6 *)filt; 
			
		log(LOG_DEBUG, 0, "Filter6 Dest Addr: %s (mask %s) Port: %d",
		    inet_ntop(AF_INET6, &sfilt6->ff_flow6.fi6_dst,
			      str1, sizeof(str1)),
		    inet_ntop(AF_INET6, &sfilt6->ff_mask6.mask6_dst,
			      str2, sizeof(str2)),
		    ntoh16(sfilt6->ff_flow6.fi6_dport));
		log(LOG_DEBUG, 0, "Filter6 Src Addr: %s (mask %s) Port: %d",
		    inet_ntop(AF_INET6, &sfilt6->ff_flow6.fi6_src,
			      str1, sizeof(str1)),
		    inet_ntop(AF_INET6, &sfilt6->ff_mask6.mask6_src,
			      str2, sizeof(str2)),
		    ntoh16(sfilt6->ff_flow6.fi6_sport));
		log(LOG_DEBUG, 0, "Filter6 Protocol: %d TCLASS 0x%x (mask 0x%x)\n",
		    sfilt6->ff_flow6.fi6_proto, sfilt6->ff_flow6.fi6_tclass,
		    sfilt6->ff_mask6.mask6_tclass);
	}
#endif /* INET6 */
}

#ifndef CBQDAEMON  /* only for RSVP */
/************************************************************************
 *
 *	Interface routines to call the CBQ kernel functions
 * 
 ************************************************************************/

/* Note: these calls use the address of the tc_t block as a parameter
 *	to carry the interface number.  This block actually contains the
 *	the other parameters too, but we leave the parameters explicit to
 *	make the code exactly parallel the Functional Specification.
 */
void
TC_init(int OIf)
{
	int i;
	static once = FALSE;

	if (once)
		return;
	once = TRUE;
	if(cbq_init() < 0) {
		log(LOG_ERR, 0, "Can't initialize CBQ!\n");
	}

        for (i = 0; i < if_num; i++) {
		cbq_ifinfo_t   *ifinfo;

		if (IsNumAPI(i))
			continue;
		if_vec[i].if_up = 0;

		for (ifinfo = cbq_iflist; ifinfo; ifinfo = ifinfo->next)
			if (strcmp(ifinfo->ifname, if_vec[i].if_name) == 0) {
				if (ifinfo->resv_class != NULL) {
					if_vec[i].if_up = 1;
				}
				else
					log(LOG_WARNING, 0,
					    "%s: No Class defined for RSVP\n",
					    ifinfo->ifname);
			}
	}
}

/*
 * TC_AddFlowspec(): Call the kernel to make reservation for a flow.
 * 	It checks admission control, and returns a handle for the
 *	reservation, or -1 if an error.
 */
u_long
TC_AddFlowspec(int OIf, FLOWSPEC *spec, SENDER_TSPEC *stspec, ADSPEC *adspp,
					int flags, FLOWSPEC **fwd_specpp)
{
	cbq_flowinfo_t 	*cbq_handle = NULL;
	tc_t		*tc_handle = NULL;

	rsvp_errno = 0;
	*fwd_specpp = NULL;

	if(IsNumAPI(OIf))
		return((u_long)NULL);

	/* Handle outgoing flow */
	rsvp_errno = 0;
	if (!if_vec[OIf].if_up) {
		rsvp_errno = Set_Errno(RSVP_Err_TC_SYS_ERROR, 0);
		return(TC_ERROR);
	}
		
	/* Create reservation control block and init */
	if((tc_handle = (tc_t*)malloc(sizeof(tc_t))) == NULL) {
		rsvp_errno = Set_Errno(RSVP_Err_TC_SYS_ERROR, 0);
		return(TC_ERROR);
	}
	tc_handle->in_if = OIf;
	tc_handle->flags = flags;
	tc_handle->cbq_handle = NULL;
	tc_handle->spec = copy_spec(spec);
	tc_handle->s_tspec = stspec ? copy_tspec(stspec) : NULL;
		
	/* Can we admit send flow? */
	switch(admit_send_flow(tc_handle, spec, stspec, flags)){
	case ADMIT_FLOW_OK:
		if((cbq_handle = cbq_addflow(tc_handle)) == NULL) {
			(void)unadmit_send_flow(tc_handle);
			rsvp_errno = Set_Errno(RSVP_Err_TC_SYS_ERROR, 0);
			goto err_ret;
		}
		tc_handle->cbq_handle = cbq_handle;
		return((u_long)tc_handle);
			
	case ADMIT_FLOW_NOSVC:
		rsvp_errno = Set_Errno(RSVP_Err_TC_ERROR,
				       RSVP_Erv_No_Serv);
		goto err_ret;
		
	case ADMIT_FLOW_NOBW:
		rsvp_errno = Set_Errno(RSVP_Err_ADMISSION, 
				       RSVP_Erv_Bandwidth);
		goto err_ret;

	case ADMIT_FLOW_SYSERR:
	default:
		rsvp_errno = Set_Errno(RSVP_Err_TC_SYS_ERROR, 0);
		goto err_ret;
	}
	
 err_ret:
    	if(tc_handle) {
    		if(tc_handle->spec) free(tc_handle->spec);
    		if(tc_handle->s_tspec) free(tc_handle->s_tspec);
    		free(tc_handle);
    	}
    	return(TC_ERROR);
}

/*
 * TC_DelFlowspec(): This routine deletes flow for specified handle
 */

int
TC_DelFlowspec(int OIf, u_long rhandle) 
{
	tc_t		*tch;

	if (rhandle == TC_ERROR)
		return TC_ERROR;
	if((tch = (tc_t*)rhandle) == NULL || IsNumAPI(OIf))
		return(TC_OK);
		
	/* Handle outgoing flow */
	unadmit_send_flow(tch);
	if(cbq_delflow(tch->cbq_handle) < 0) {
		if(!rsvp_errno) 
			rsvp_errno = Set_Errno(RSVP_Err_TC_SYS_ERROR, 0);
		return TC_ERROR;
	}
	if(tch->spec) free(tch->spec);
	if(tch->s_tspec) free(tch->s_tspec);
	free(tch);
	return TC_OK;
}

/*
 * TC_ModFlowspec(): Modifies a flowspec of a given flow.
 */
int
TC_ModFlowspec(int OIf, u_long rhandle,
	       FLOWSPEC *specp, SENDER_TSPEC *stspecp, ADSPEC *adspecp,
	       int flags, FLOWSPEC **fwd_specpp)
{
	tc_t		*tc_handle;
	tc_t		tc_ohandle;
	cbq_flowinfo_t	*cbq_handle;
	cbq_ifinfo_t	*ifinfo;

	rsvp_errno = 0;
	*fwd_specpp = NULL;

	if((tc_handle = (tc_t*)rhandle) == NULL || IsNumAPI(OIf))
		return (TC_OK);

	/* Handle outgoing flow */
	cbq_handle = tc_handle->cbq_handle;
	ifinfo = cbq_handle->clinfo->ifinfo;

	/*
	 * first, check if we can admit this update.
	 * save old flowspec in case of admission failure.
	 */
	tc_ohandle = *tc_handle;
	unadmit_send_flow(tc_handle);

	tc_handle->flags = flags;
	tc_handle->spec = copy_spec(specp);
	tc_handle->s_tspec = stspecp ? copy_tspec(stspecp) : NULL;
	
	/* Can we admit new send flow? */
	switch(admit_send_flow(tc_handle, specp, stspecp, flags)){
	case ADMIT_FLOW_OK:
		break;
	case ADMIT_FLOW_NOSVC:
		rsvp_errno = Set_Errno(RSVP_Err_TC_ERROR, RSVP_Erv_No_Serv);
		break;
	case ADMIT_FLOW_NOBW:
		rsvp_errno = Set_Errno(RSVP_Err_ADMISSION, RSVP_Erv_Bandwidth);
		break;
	case ADMIT_FLOW_SYSERR:
	default:
		rsvp_errno = Set_Errno(RSVP_Err_TC_SYS_ERROR, 0);
		break;
	}

	if (rsvp_errno) {
		/* can't update.  restore the original flowspec */
		if(tc_handle->spec) free(tc_handle->spec);
		if(tc_handle->s_tspec) free(tc_handle->s_tspec);

		tc_handle->flags = tc_ohandle.flags;
		tc_handle->spec = tc_ohandle.spec;
		tc_handle->s_tspec = tc_ohandle.s_tspec;
		if (admit_send_flow(tc_handle, tc_handle->spec,
				    tc_handle->s_tspec, tc_handle->flags)
		    != ADMIT_FLOW_OK) {
			/* ick! can't restore the original flow! */
			log(LOG_ERR, 0, "can't restore the original flow!\n");
		}
		return (TC_ERROR);
	}

	/*
	 * now we have decided to commit the update.
	 * errors should not happen from now on, since we are no longer
	 * able to restore the original flowspec.
	 */
	if(tc_ohandle.spec) free(tc_ohandle.spec);
	if(tc_ohandle.s_tspec) free(tc_ohandle.s_tspec);

	if((cbq_handle = cbq_modifyflow(tc_handle)) == NULL) {
		log(LOG_ERR, errno, "can't modify flowspec!");
		(void)unadmit_send_flow(tc_handle);
		tc_handle->cbq_handle = NULL;
		rsvp_errno = Set_Errno(RSVP_Err_TC_SYS_ERROR, 0);
		return (TC_ERROR);
	}
	tc_handle->cbq_handle = cbq_handle;

	return (TC_OK);
}

/*
 * TC_AddFilter(): Adds a filter for an existing flow.
 *
 *	Returns fhandle or TC_ERROR.
 */
u_long
TC_AddFilter(int OIf, u_long rhandle, Session *dest, FILTER_SPEC *filtp)
{
	struct flow_filter cfilter;
	tc_t		*tch = (tc_t*)rhandle;
	cbq_fltrinfo_t	*fh;

	if(tch == NULL || IsNumAPI(tch->in_if) || !if_vec[tch->in_if].if_up) {
		/* CBQ does not create kernel state for incoming flows */
		return TC_OK;
	}

	rsvp_errno = 0;
		
	/* Add or modify the filter */
	if(rsvp2cbq_filter(filtp, dest, &cfilter) < 0 ||
	   (fh = cbq_addfilter(tch->cbq_handle, &cfilter)) == NULL) {
		if(!rsvp_errno) 
			rsvp_errno = RSVP_Err_TC_SYS_ERROR;
		return TC_ERROR;
	}
	return (u_long)fh;
}


/*
 * TC_DelFilter(): Deletes existing filter.
 */
int
TC_DelFilter(int OIf, u_long fhandle)
{
	cbq_fltrinfo_t *fh = (cbq_fltrinfo_t *)fhandle;

	if(fh == NULL || IsNumAPI(OIf) || !if_vec[OIf].if_up)
		return TC_OK;

	rsvp_errno = 0;
	if(cbq_delfilter(fh) < 0) {
		if(!rsvp_errno) 
			rsvp_errno = RSVP_Err_TC_SYS_ERROR;
		return TC_ERROR;
	}
	return TC_OK;
}


#define In_Obj(x, y) ((Object_header *)(x) <= Next_Object((Object_header *)(y)))

/*
 * TC_Advertise(): Given existing OPWA ADSPEC, return a new updated object.
 */
ADSPEC *
TC_Advertise(int OIf, ADSPEC * old_asp, int flags)
{
	ADSPEC		*new_asp = copy_adspec(old_asp);
	/*	We know we will not expand the ADSPEC, so start by
	 *	just making a straight copy.
	 */
	IS_main_hdr_t	*mhp = (IS_main_hdr_t *) Obj_data(new_asp);
	genparm_parms_t *gpp = (genparm_parms_t *)(mhp +1);
	IS_serv_hdr_t	*shp, *lastshp;
	float32_t	hop_bw;
	u_int32_t	hop_latency, hop_mtu;
	cbq_ifinfo_t	*ifinfo;

	if (if_vec[OIf].if_up == 0 || (flags & ADVERTF_NonRSVP)) {
		Set_Break_Bit(&gpp->gen_parm_hdr);
	}
	else {
		/* get interface info, to over-ride interface defaults */
		if ((ifinfo = ifname2ifinfo(if_vec[OIf].if_name)) != NULL)
			if_vec[OIf].if_path_bw =
				ifinfo->resv_class->acinfo.bandwidth
				- ifinfo->resv_class->acinfo.allocated;
	}
	
	gpp->gen_parm_hopcnt++;

	hop_bw = (if_vec[OIf].if_path_bw)?if_vec[OIf].if_path_bw:
							(TC_DFLT_PATH_BW);
	gpp->gen_parm_path_bw = MIN(gpp->gen_parm_path_bw, hop_bw);

	hop_latency = (if_vec[OIf].if_min_latency)?if_vec[OIf].if_min_latency:
							TC_DFLT_MIN_LATENCY;
	gpp->gen_parm_min_latency = MIN(gpp->gen_parm_min_latency, hop_latency);
	
	hop_mtu = (if_vec[OIf].if_path_mtu)?if_vec[OIf].if_path_mtu: 
							TC_DFLT_MTU;
	gpp->gen_parm_composed_MTU = MIN(gpp->gen_parm_composed_MTU, hop_mtu);
	
	/* If we are not capable of traffic control for a service,
	 * set Break bit in the service.  We DO NOT set the Break
	 * bit in the generic parameters, because we DO understand
	 * ADSPECs, unless previous hop was not RSVP-capable.
	 */
	shp = Next_Serv_Hdr((IS_serv_hdr_t *)gpp);
	lastshp  = (IS_serv_hdr_t *) Next_Main_Hdr(mhp);
	if (!In_Obj(lastshp, new_asp)) {
		/* Internal length falls outside object...
		 *  Increment error counter
		 */
		free(new_asp);
		return(NULL);
	}
	while (shp < lastshp) {
		switch(shp->issh_service) {
		case CONTROLLED_LOAD_SERV:
			if (if_vec[OIf].if_up)
				break;
			/* fall into... */
		default:
			Set_Break_Bit(shp);
			break;
		}
		shp = Next_Serv_Hdr(shp);
	}
	return(new_asp);
}


static cbq_flowinfo_t *
cbq_addflow(tc_t *tch)
{
	IS_specbody_t		*z_specp;
	IS_serv_hdr_t		*sp;
	cbq_ifinfo_t  		*ifinfo;
	cbq_acinfo_t   		*acinfo;
	cbq_classinfo_t 	*clinfo;
	u_long          	cl_handle;
	cbq_flowinfo_t 		*new_fl;
	u_int			bandwidth;

	assert(tch);
	assert(Obj_CType(tch->spec) == ctype_FLOWSPEC_Intserv0);
	z_specp = &(tch->spec->flow_body);
	sp = (IS_serv_hdr_t *) &z_specp->spec_u;
	switch(sp->issh_service) {
	case GUARANTEED_SERV:
		bandwidth = z_specp->spec_u.G_spec.Gspec_r * 8;
		break;
	case CONTROLLED_LOAD_SERV:
		bandwidth = z_specp->spec_u.CL_spec.CLspec_r * 8;
		break;
	}

	/* get interface info */
	if ((ifinfo = ifname2ifinfo(if_vec[tch->in_if].if_name)) == NULL)
		return NULL;
	if ((clinfo = ifinfo->resv_class) == NULL)
		return NULL;
	acinfo = &(ifinfo->resv_class->acinfo);
	
	if ((new_fl = calloc(1, sizeof(cbq_flowinfo_t))) == NULL)
		return NULL;
	
	/* create a leaf class */
	cl_handle = cbq_create_class(ifinfo, clinfo->handle, clinfo->handle,
	        acinfo->pri, acinfo->flags, bandwidth, acinfo->maxdelay,
		acinfo->maxburst, acinfo->minburst, acinfo->av_pkt_size,
	        acinfo->max_pkt_size);
	if (cl_handle == NULL_CLASS_HANDLE) {
		free(new_fl);
		return NULL;
	}

	new_fl->clinfo = clinfo;
	new_fl->bandwidth = bandwidth;
	new_fl->handle = cl_handle;
	new_fl->next = acinfo->flow_list;
	acinfo->flow_list = new_fl;
	return new_fl;
}

static cbq_flowinfo_t *
cbq_modifyflow(tc_t *tch)
{
	IS_specbody_t		*z_specp;
	IS_serv_hdr_t		*sp;
	cbq_ifinfo_t  		*ifinfo;
	cbq_acinfo_t   		*acinfo;
	cbq_classinfo_t 	*clinfo;
	u_long          	cl_handle;
	cbq_flowinfo_t 		*cbq_handle;
	u_int			bandwidth;

	assert(tch);
	assert(Obj_CType(tch->spec) == ctype_FLOWSPEC_Intserv0);
	z_specp = &(tch->spec->flow_body);
	sp = (IS_serv_hdr_t *) &z_specp->spec_u;
	switch(sp->issh_service) {
	case GUARANTEED_SERV:
		bandwidth = z_specp->spec_u.G_spec.Gspec_r * 8;
		break;
	case CONTROLLED_LOAD_SERV:
		bandwidth = z_specp->spec_u.CL_spec.CLspec_r * 8;
		break;
	}

	/* get interface info */
	if ((ifinfo = ifname2ifinfo(if_vec[tch->in_if].if_name)) == NULL)
		return NULL;
	if ((clinfo = ifinfo->resv_class) == NULL)
		return NULL;
	acinfo = &(ifinfo->resv_class->acinfo);
	
	if ((cbq_handle = tch->cbq_handle) == NULL)
		return NULL;
	
	/* modify the specified class */
	cl_handle = cbq_modify_class(ifinfo, cbq_handle->handle,
		clinfo->handle, clinfo->handle,
	        acinfo->pri, acinfo->flags, bandwidth, acinfo->maxdelay,
		acinfo->maxburst, acinfo->minburst, acinfo->av_pkt_size,
	        acinfo->max_pkt_size);
	if (cl_handle == NULL_CLASS_HANDLE)
		return NULL;

	cbq_handle->bandwidth = bandwidth;
	cbq_handle->handle = cl_handle;
	return (cbq_handle);
}

static int
cbq_delflow(cbq_flowinfo_t * flow_info)
{
	cbq_classinfo_t *clinfo = flow_info->clinfo;
	cbq_ifinfo_t	*ifinfo = clinfo->ifinfo;
	struct cbq_delete_class	class_delete;
	struct cbq_flowinfo	*flow_list;

	/* delete any outstanding filters on class for this flow */
	if (cbq_delfilter_all(flow_info) < 0)
		return (-1);

	if (flow_info->handle != NULL_CLASS_HANDLE) {
		/* delete class from kernel */
		memset((void *) &class_delete, 0, sizeof(class_delete));
		strncpy(class_delete.cbq_iface.cbq_ifacename,
			ifinfo->ifname, IFNAMSIZ);
		class_delete.cbq_class_handle = flow_info->handle;
		if (ifinfo->use_cbq &&
		    ioctl(cbqfd, CBQ_DEL_CLASS, &class_delete) < 0) {
			log(LOG_ERR, errno, "ioctl CBQ_DEL_CLASS");
			return (-1);
		}
	}

	/* remove from list of flows for this reservation class */
	if (clinfo->acinfo.flow_list == flow_info) {
		clinfo->acinfo.flow_list = flow_info->next;
	} else {
		flow_list = clinfo->acinfo.flow_list;
		while (flow_list->next) {
			if (flow_list->next == flow_info) {
				flow_list->next = flow_info->next;
				break;
			}
			flow_list = flow_list->next;
		}
	}
	free(flow_info);
	return 0;
}

static cbq_fltrinfo_t*
cbq_addfilter(cbq_flowinfo_t *flinfo, struct flow_filter *cfilter)
{
	cbq_ifinfo_t   *ifinfo = flinfo->clinfo->ifinfo;
	struct cbq_add_filter fltr_add;
	cbq_fltrinfo_t *fltr_info;

	if ((fltr_info = calloc(1, sizeof(cbq_fltrinfo_t))) == NULL)
		return NULL;

	memset((void *) &fltr_add, 0, sizeof(fltr_add));
	strncpy(fltr_add.cbq_iface.cbq_ifacename, ifinfo->ifname, IFNAMSIZ);
	fltr_add.cbq_class_handle = flinfo->handle;
	fltr_add.cbq_filter = *cfilter;

	if (IsDebug(DEBUG_CBQ)) {
		log(LOG_DEBUG, 0, "%s: CBQ_ADD_FILTER, class.%lx\n",
		    ifinfo->ifname, fltr_add.cbq_class_handle);
		pr_filter(cfilter);
	}

	if (ifinfo->use_cbq &&
	    ioctl(cbqfd, CBQ_ADD_FILTER, &fltr_add) < 0) {
		log(LOG_ERR, errno, "ioctl CBQ_ADD_FILTER");
		free(fltr_info);
		return NULL;
	}

	fltr_info->handle = fltr_add.cbq_filter_handle;
	fltr_info->filter = *cfilter;
	fltr_info->flow = flinfo;

	fltr_info->next = flinfo->fltr_list;
	flinfo->fltr_list = fltr_info;

	return (fltr_info);
}

static int
cbq_delfilter_all(cbq_flowinfo_t * flow_info)
{
	while (flow_info->fltr_list != NULL)
		if (cbq_delfilter(flow_info->fltr_list) < 0)
			return (-1);
	return (0);
}

static int 
cbq_delfilter(cbq_fltrinfo_t *fltr)
{
	struct cbq_delete_filter fltr_delete;
	cbq_flowinfo_t *flinfo = fltr->flow;
	cbq_ifinfo_t   *ifinfo = flinfo->clinfo->ifinfo;
	cbq_fltrinfo_t *fltr_prev;
	
	memset((void *) &fltr_delete, 0, sizeof(fltr_delete));
	strncpy(fltr_delete.cbq_iface.cbq_ifacename, ifinfo->ifname, IFNAMSIZ);
	fltr_delete.cbq_filter_handle = fltr->handle;

	if (IsDebug(DEBUG_CBQ)) {
		log(LOG_DEBUG, 0, "%s: CBQ_DEL_FILTER\n", ifinfo->ifname);
		pr_filter(&fltr->filter);
	}

	if (ifinfo->use_cbq &&
	    ioctl(cbqfd, CBQ_DEL_FILTER, &fltr_delete) < 0) {
		log(LOG_ERR, errno, "ioctl CBQ_DEL_FILTER");
		return (-1);
	}

	/* remove from the filter list */
	if (flinfo->fltr_list == fltr)
		flinfo->fltr_list = fltr->next;
	else {
		fltr_prev = flinfo->fltr_list;
		while (fltr_prev->next != NULL) {
			if (fltr_prev->next == fltr) {
				fltr_prev->next = fltr->next;
				break;
			}
			fltr_prev = fltr_prev->next;
		}
	}
	free(fltr);
	return (0);
}

#ifndef filt4_srcaddr
/* for isi rsvp rel4.1 compatibility */
#define filt4			filt_u.filt_ipv4
#define filt4_srcaddr		filt4.filt_ipaddr
#define filt4_srcport		filt4.filt_port
#endif

static int
rsvp2cbq_filter(FilterSpec *filtp, Session *dest, struct flow_filter *flp)
{
#ifdef USE_IPV6			
	struct flow_filter6 *fl6p;
#endif

	memset(flp, 0, sizeof(struct flow_filter));

	switch (Obj_CType(dest->d_session)) {
	case ctype_SESSION_ipv4:
	case ctype_SESSION_ipv4GPI:
		flp->ff_flow.fi_family = AF_INET;
		flp->ff_flow.fi_dst  = dest->d_session->sess4_addr;
		flp->ff_mask.mask_dst.s_addr  = 0xffffffff;
		flp->ff_flow.fi_dport  = dest->d_session->sess4_port;
		flp->ff_flow.fi_proto  = dest->d_session->sess4_prot;

		if (filtp == NULL) {
			/* IPv4 wildcard filter */
			flp->ff_flow.fi_src.s_addr = 0;
			flp->ff_flow.fi_sport = 0;
			
		} else switch (Obj_CType(filtp)) {
		case ctype_FILTER_SPEC_ipv4:	/* IPv4 FILTER_SPEC */
			flp->ff_flow.fi_src = filtp->filt4_srcaddr;
			flp->ff_mask.mask_src.s_addr = 0xffffffff;
			flp->ff_flow.fi_sport = filtp->filt4_srcport;
			break;

		case ctype_FILTER_SPEC_ipv4GPI:	/* IPv4/GPI FILTER_SPEC */
			flp->ff_flow.fi_dport  = 0;/* we don't need dst port */
			flp->ff_flow.fi_src = filtp->filtgpi4_srcaddr;
			flp->ff_mask.mask_src.s_addr = 0xffffffff;
			flp->ff_flow.fi_gpi = filtp->filtgpi4_srcgpi;
			break;
			
		default:
			rsvp_errno = Set_Errno(RSVP_Err_UNKNOWN_CTYPE,
					       Obj_CType(filtp));
			return (-1);
		}
		
		break;

#ifdef USE_IPV6			
	case ctype_SESSION_ipv6:
	case ctype_SESSION_ipv6GPI:
		fl6p = (struct flow_filter6 *)flp;
		fl6p->ff_flow6.fi6_family = AF_INET6;
		fl6p->ff_flow6.fi6_dst = dest->d_session->sess6_addr;
		memset(&fl6p->ff_mask6.mask6_dst, 0xff,
		       sizeof(struct in6_addr));
		fl6p->ff_flow6.fi6_dport = dest->d_session->sess6_port;
		fl6p->ff_flow6.fi6_proto = dest->d_session->sess6_prot;

		if (filtp == NULL) {
			/* IPv6 wildcard */
			memset(&fl6p->ff_flow6.fi6_src, 0,
			       sizeof(struct in6_addr));
			fl6p->ff_flow6.fi6_sport = 0;
			
		} else switch (Obj_CType(filtp)) {
		case ctype_FILTER_SPEC_ipv6:	/* IP6 FILTER_SPEC */
			fl6p->ff_flow6.fi6_src = filtp->filt6_srcaddr;
			memset(&fl6p->ff_mask6.mask6_src, 0xff,
			       sizeof(struct in6_addr));
			fl6p->ff_flow6.fi6_sport = filtp->filt6_srcport;
			break;
			
#ifdef filtfl6_srcaddr
		case ctype_FILTER_SPEC_ipv6FL: /* IP6 Flow-label FILTER_SPEC */
			fl6p->ff_flow6.fi6_dport = 0; /* we don't need this */
			fl6p->ff_flow6.fi6_src = filtp->filtfl6_srcaddr;
			memset(&fl6p->ff_mask6.mask6_src, 0xff,
			       sizeof(struct in6_addr));
			fl6p->ff_flow6.fi6_flowlabel = filtp->filtfl6_srcfl
				& htonl(0x000fffff);
			break;
#endif

		case ctype_FILTER_SPEC_ipv6GPI:	/* IPv6/GPI FILTER_SPEC */
			fl6p->ff_flow6.fi6_dport = 0; /* we don't need this */
			fl6p->ff_flow6.fi6_src = filtp->filtgpi6_srcaddr;
			memset(&fl6p->ff_mask6.mask6_src, 0xff,
			       sizeof(struct in6_addr));
			fl6p->ff_flow6.fi6_gpi = filtp->filtgpi6_srcgpi;
			break;

		default:
			rsvp_errno = Set_Errno(RSVP_Err_UNKNOWN_CTYPE,
					       Obj_CType(filtp));
			return (-1);
		}

		break;
#endif /* USE_IPV6 */

	default:
		rsvp_errno = Set_Errno(RSVP_Err_UNKNOWN_CTYPE,
				       Obj_CType(dest->d_session));
		return (-1);
		break;
	}
	
	return (0);
}

static int
admit_send_flow(tc_t *tch, FLOWSPEC *spec, SENDER_TSPEC *stspec, int flags)
{
	cbq_ifinfo_t		*ifinfo;
	cbq_acinfo_t		*acinfo;
	cbq_classinfo_t 	*clinfo;
	IS_specbody_t		*z_specp;
	IS_serv_hdr_t		*sp;
	u_int			req_bandwidth;

	assert(tch);
	
	if ((ifinfo = ifname2ifinfo(if_vec[tch->in_if].if_name)) == NULL)
		return ADMIT_FLOW_SYSERR;

	/* Get relevant information from flow spec */
	if(Obj_CType(tch->spec) != ctype_FLOWSPEC_Intserv0) {
		return ADMIT_FLOW_NOSVC;
	}
	z_specp = &(tch->spec->flow_body);

	/* Check against interface first */
	sp = (IS_serv_hdr_t *) &z_specp->spec_u;
	switch(sp->issh_service) {
	case GUARANTEED_SERV:
		if(!Allow_Guaranteed)
			return ADMIT_FLOW_NOSVC;

		req_bandwidth = z_specp->spec_u.G_spec.Gspec_r * 8;
		break;

	case CONTROLLED_LOAD_SERV:
		req_bandwidth =  z_specp->spec_u.CL_spec.CLspec_r * 8;
		break;

	default:
		return ADMIT_FLOW_NOSVC;
	}

	if ((clinfo = ifinfo->resv_class) == NULL)
		return ADMIT_FLOW_SYSERR;
	acinfo = &(ifinfo->resv_class->acinfo);
	if ((acinfo->allocated + req_bandwidth) > acinfo->bandwidth)
		return ADMIT_FLOW_NOBW;
	
	/* All checks passed */
	acinfo->allocated += req_bandwidth;
	acinfo->nflows++;
	return ADMIT_FLOW_OK;
}

static void
unadmit_send_flow(tc_t *tch)
{
	cbq_ifinfo_t		*ifinfo;
	cbq_acinfo_t		*acinfo;
	cbq_classinfo_t 	*clinfo;
	IS_specbody_t		*z_specp;
	IS_serv_hdr_t		*sp;
	u_int			req_bandwidth;

	assert(tch);

	if ((ifinfo = ifname2ifinfo(if_vec[tch->in_if].if_name)) == NULL)
		return;

	assert(Obj_CType(tch->spec) == ctype_FLOWSPEC_Intserv0);
	z_specp = &(tch->spec->flow_body);
	sp = (IS_serv_hdr_t *) &z_specp->spec_u;
	switch(sp->issh_service) {
	case GUARANTEED_SERV:
		if(!Allow_Guaranteed)
			return;

		req_bandwidth = z_specp->spec_u.G_spec.Gspec_r * 8;
		break;

	case CONTROLLED_LOAD_SERV:
		req_bandwidth = z_specp->spec_u.CL_spec.CLspec_r * 8;
		break;

	default:
		return;
	}
	clinfo = ifinfo->resv_class; 
	acinfo = &(ifinfo->resv_class->acinfo);
	acinfo->allocated -= req_bandwidth;
	acinfo->nflows--;
}

#endif /* !CBQDAEMON */


