/* $Id: fifoqd.c,v 0.4 1999/12/28 11:57:21 kjc Exp kjc $ */
/*
 * Copyright (C) 1997
 *	Sony Computer Science Laboratory Inc.  All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY SONY CSL AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL SONY CSL OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 */

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>
#include <errno.h>
#include <fcntl.h>
#include <signal.h>
#include <syslog.h>

#include <sys/param.h>
#include <sys/ioctl.h>
#include <sys/socket.h>

#include <net/if.h>

#include <altq/altq.h>
#include <altq/altq_fifoq.h>

#define FIFOQ_DEV	"/dev/altq/fifoq"

int fd = -1;
int debug = 0;
int maxqlen = 0;

static void usage(void)
{
    fprintf(stderr, "usage: fifoqd [-d] [-l maxqlen] if_name\n");
    fprintf(stderr, "		-d: debug\n");
    fprintf(stderr, "		-l maxqlen: set max queue length\n");
    exit(1);
}

static void
sigint_handler(int sig)
{
    /* do nothing */
}

int
main(argc, argv)
	int argc;
	char *argv[];
{
    int ch;
    struct fifoq_interface iface;
    char *if_name;

    while ((ch = getopt(argc, argv, "dl:")) != EOF) {
	switch (ch) {
	case 'd':
	    debug = 1;
	    break;
	case 'l':
	    maxqlen = atoi(optarg);
	    break;
	}
    }
    argc -= optind;
    argv += optind;
    if (argc > 0) {
	if_name = argv[0];
    }
    else {
	if (argc == 0)
	    fprintf(stderr, "missing interface name\n");
	usage();
    }

    signal(SIGINT, sigint_handler);
    signal(SIGTERM, sigint_handler);

    openlog("fifoqd", (debug ? (LOG_PID|LOG_PERROR) : LOG_PID), LOG_DAEMON);

    if (!debug)
	daemon(0, 0);

    if((fd = open(FIFOQ_DEV, O_RDWR)) < 0) {
	syslog(LOG_ERR, "can't open %s: %m", FIFOQ_DEV);
	exit(1);
    }

    if (debug) {
	printf("fifoqd on interface %s\n", if_name);
	if (maxqlen > 0)
	    printf("  set queue limit to %d\n", maxqlen);
    }

    iface.fifoq_ifname[IFNAMSIZ-1] = '\0';
    strncpy(iface.fifoq_ifname, if_name, IFNAMSIZ-1);
    
    if (ioctl(fd, FIFOQ_IF_ATTACH, &iface) < 0) {
	syslog(LOG_ERR, "FIFOQ_IF_ATTACH for %s: %m", if_name);
	exit(1);
    }

    if (maxqlen > 0) {
	struct fifoq_conf conf;

	conf.iface = iface;
	conf.fifoq_limit = maxqlen;
	if (ioctl(fd, FIFOQ_CONFIG, &conf) < 0) {
	    syslog(LOG_ERR, "FIFOQ_CONFIG: %m");
	    exit(1);
	}
    }

    if (ioctl(fd, FIFOQ_ENABLE, &iface) < 0) {
	syslog(LOG_ERR, "FIFOQ_ENABLE: %m");
	exit(1);
    }

    /* wait for signals and then terminate */
    pause();

    if (debug)
	printf("Got signal.  Exiting...\n");

    if (ioctl(fd, FIFOQ_DISABLE, &iface) < 0)
	syslog(LOG_ERR, "FIFOQ_DISABLE: %m");

    if (ioctl(fd, FIFOQ_IF_DETACH, &iface) < 0)
	syslog(LOG_ERR, "FIFOQ_IF_DETACH: %m");

    close(fd);
    closelog();

    return 0;
}
